/*
 * Copyright 2005, 2006 University of Leiden.
 *
 * This file is part of MIA+EWS.
 *
 * MIA+EWS is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * MIA+EWS is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with MIA; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <math.h>
#include "fitsio.h"
#include "imaging_fdata.h"
#include "delay.h"
#include "dispersion.h"
#define clight  2.9979250e8 /*m per sec */
#define twopi   6.2831853072
#define radeg  57.29578
#define MGRAD 8 /* max number of trial gradients */
#define PGRAD .5  /*radians/sigma*/
#define NDEBIAS 1 
#define SMOOTHSIGMA 0.2 /* sec */
void Usage() {
   printf("   oirRotateGroupDelay infile delayfile outfile [-smooth smooth] [-frames]\n");
   printf("         [-ngrad ngrad] [-phase phasefile] [-mask maskfile][-nophase] [-nogroup] [-noopd] \n");
   printf("      infile = file name of output of oirFormFringes\n");
   printf("      delayfile = file name of output of oirGroupDelay\n");
   printf("      outfile = name of output file\n");
   printf("      smooth (optional) = size of gaussian 1-sigma smoothing interval before \n");
   printf("         determining water vapor dispersion phase\n");
   printf("      frames (optional):if specified smooth is specified in frames rather than seconds (default)\n");
   printf("      ngrad (optional) = number of trial phase gradients (wrt time) to try\n");
   printf("         default=0(assume constant phase per smoothing interval)\n");
   printf("      phasefile = name of file containing previous solved phase\n");
   printf("         optional; if not found phases will be found from data\n");
   printf("         in seconds unless -frames is specified; default=0.2sec\n");
   printf("      maskfile (optional) = use a mask to suppress noisy regions before determing\n");
   printf("         water vapor dispersion phase; default: no masking\n");
   printf("      nophase (optional):if specified do not perform water vapor phase correction\n");
   printf("      noopd (optional):  if specified do not remove instrumental OPD\n");
   printf("      nogroup (optional):if specified do not remove atmospheric group OPD\n\n");
}
/*
 *          get two mask images from file
 */
void getMaskImages(char *maskFileName, float *maskImages[2], 
   int *nx, int *ny, int *ierr) {

   fitsfile *maskFile;
   imaging_detector_table_def *DetectorDef;
   imaging_fdata_table_def    *idata;
   imaging_fdata_row *inData, *inData1;

   int i, ix, npix;
   int iRegion;
   float x, sigma;

   fits_open_file (&maskFile, maskFileName, READONLY, ierr);
   if (*ierr != 0) {
      printf("fits openMaskfile of %s gave error %i",maskFileName, *ierr);
      printf("%s \n",strerror(errno));
      return;
   }
   *ierr = 0;
   DetectorDef = read_imaging_detector_table(maskFile, ierr);
   if (*ierr !=0) {
      printf("error %i opening detector table in %s\n",
         *ierr, maskFileName);
      printf("%s \n",strerror(errno));
      return ;
   }

   idata = open_imaging_fdata_table(maskFile, ierr);
   if (*ierr !=0) {
      printf("error %i opening maskFile data in file %s\n",
         *ierr, maskFileName);
      printf("%s \n",strerror(errno));
      return;
   }
   inData = make_imaging_fdata_row(idata);
   i = read_imaging_fdata_table(maskFile, idata, inData, 1, ierr);
   if (*ierr !=0) {
      printf("error %i reading maskFile data in file %s\n",
         *ierr, maskFileName);
      printf("%s \n",strerror(errno));
      return;
   }
   *nx = DetectorDef->naxis[0][0];
   *ny = DetectorDef->naxis[0][1];
/*
           allocate array and copy data
*/
   npix = (*nx) * (*ny);
   for (iRegion=0;iRegion<2;iRegion++) {
      maskImages[iRegion] = (float *)malloc(npix*sizeof(float));
      for (i=0;i<npix;i++) maskImages[iRegion][i] = inData->data[iRegion][i];
   }
    i= close_imaging_fdata_table(maskFile, idata, ierr);
    kill_imaging_fdata_row(inData);
   return;
} /* end of getMaskImages */

/*
 *       for each valid frame sum up all the complex flux values so
 *       we can later figure out their average phases
*/
void sumFluxes(int nRegion, imaging_fdata_row **outDataArray, int nRow,
   int nx, int ny, int isMask, float** maskImages, float *totalFlux) {

   int iRow, iRegion, ix, iy;
   int y1, y2;
   float rWeight[2];
    
   rWeight[0] =  1.;
   rWeight[1] = -1.;

   if (ny == 1) { y1 = 0; y2 = 0;} else {y1 = 10; y2 = 30;}
   if (y2 > ny) y2 = ny-1;
   if (isMask) {
      for (iRow=0;iRow<nRow;iRow++) 
      for (iRegion=0; iRegion<nRegion; iRegion++)
      for (iy=y1;iy<=y2; iy++) 
      for (ix=0;ix<nx;ix++) {
         totalFlux[2*iRow] = totalFlux[2*iRow] + 
            maskImages[iRegion][iy*nx+ix]*rWeight[iRegion]*
            outDataArray[iRow]->data[iRegion][2*ix+2*nx*iy];
         totalFlux[2*iRow+1] = totalFlux[2*iRow+1] + 
            maskImages[iRegion][iy*nx+ix]*rWeight[iRegion]*
            outDataArray[iRow]->data[iRegion][2*ix+2*nx*iy+1];
      }

   } else {
      for (iRow=0;iRow<nRow;iRow++) 
      for (iRegion=0; iRegion<nRegion; iRegion++)
      for (iy=y1;iy<=y2; iy++) 
      for (ix=0;ix<nx;ix++) {
         totalFlux[2*iRow] = totalFlux[2*iRow] + 
            rWeight[iRegion]* outDataArray[iRow]->data[iRegion][2*ix+2*nx*iy];
         totalFlux[2*iRow+1] = totalFlux[2*iRow+1] + 
            rWeight[iRegion]* outDataArray[iRow]->data[iRegion][2*ix+2*nx*iy+1];
       }
   }
   return;
}

/* ************************************************************************
 * Find the index and value of the maximum and mininum of a float array
 * Inputs:
 *         a    array of input values 
 *         n    number of input values
 * Outputs:
 *         *min value of mininum of array
 *         *max value of maximum of array
 *         *imin position (0-rel) of minimum value in array
 *         *imax position (0-rel) of maximum value in array
 *   
 *********************************************************************** */
void mmax(float *a, int n, float *min, float *max, int *imin, int *imax) {
   int i,j,k;
   float in,ax;
   j=0;
   k=0;
   in = a[0];
   ax = in;
   for (i=0;i<n;i++){
      if (a[i] > ax) {
         ax = a[i];
         j  = i;
      } 
      if (a[i] < in) {
         in = a[i];
         k  = i;
      } 
   }
   *imin = k;
   *imax = j;
   *min  = in;
   *max  = ax;
   return;
}
/*
 *              given an estimate of the average phase for a single
 *              frame, remove it
*/
void rotateDataPhase(int nRegion, float **data, int nx, int ny, float phase) {
   float c,s,w;
   int i,j;
   int iRegion;

   c = cos(phase);
   s = sin(phase);
   for (iRegion=0;iRegion<nRegion;iRegion++) 
   for (j=0;j<ny;j++) for (i=0;i<nx; i++) {
      w  = c*data[iRegion][2*i+2*j*nx]   
         - s*data[iRegion][2*i+2*j*nx+1];
      data[iRegion][2*i+2*j*nx+1] = c*data[iRegion][2*i+2*j*nx+1] 
                                  + s*data[iRegion][2*i+2*j*nx];
      data[iRegion][2*i+2*j*nx]   = w;
   }
   return;
} /* end of rotateData Phase */


/*
 *   try to estimate average phase (due to water vapor dispersion)
 *   from the data itself
 */
void findRemovePhases(int nRegion, imaging_fdata_row **dataArray, 
   int nRow, int nx, int ny, float *phase, float sigma, 
   char *maskFileName, int nGrad, int *ierr) {

   float *totalFlux;
   int    isMask;
   float *maskImages[2];
   int    xMask, yMask;
   int    iRegion, ix, iy, iRow;
   int    xmin, xmax, rmin, rmax;
   int    iSigma;
   float  c,s, cTotal, sTotal, w;
   float  cTotalC, sTotalC;

   float  phasePerFrame, dPhase;
   float  pkSumFlux[2*MGRAD+1];
   int    iGrad;

float    tmpspec[640];
int      idb;
/*
 *        if mask specified, get it
 */
   if (maskFileName == NULL) isMask = 0;
   else {
      getMaskImages(maskFileName, maskImages, &xMask, &yMask, ierr);
      if (xMask != nx) *ierr = 2; /* different length windows */
      isMask = (*ierr == 0);
   }
   if (isMask) printf("   using maskfile %s \n", maskFileName);
   else if (*ierr !=0) {
      fprintf(stderr, "maskfile %s gave error %i\n",maskFileName, *ierr);
      return ;
   }
/*
 *       compress mask if data is 1-dimensional
*/
   if ((isMask) && (ny == 1)) for (iRegion=0;iRegion<nRegion;iRegion++) 
      for (ix=0;ix<nx;ix++) for (iy=1;iy<yMask;iy++) maskImages[iRegion][ix]=
         maskImages[iRegion][ix] + maskImages[iRegion][ix+iy*nx];

   totalFlux = (float *)malloc(2*nRow*sizeof(float));
   memset(totalFlux, 0, 2*nRow*sizeof(float));
   sumFluxes(nRegion, dataArray, nRow, nx, ny, isMask, 
         maskImages, totalFlux);
/*    consider that the phase may change a little per frame */
   phasePerFrame = PGRAD/(2*sigma+1.);

/* dummest possible phase fit */
   iSigma = (int)(sigma+.5);
   for (iRow=0; iRow<nRow; iRow++) {
      rmin = iRow - iSigma; if (rmin <0) rmin=0;
      rmax = iRow + iSigma; if (rmax >nRow) rmax=nRow;

      for (iGrad=-nGrad;iGrad<=nGrad;iGrad++) {
         cTotal = 0.;
         sTotal = 0.;
         for (ix=rmin; ix<rmax ;ix++) {
            dPhase = iGrad * phasePerFrame * (ix-iRow);
            s      = sin(dPhase);
            c      = cos(dPhase);
            cTotal = cTotal + totalFlux[2*ix]*c - totalFlux[2*ix+1]*s;
            sTotal = sTotal + totalFlux[2*ix]*s + totalFlux[2*ix+1]*c;
         } 
         pkSumFlux[iGrad+nGrad] = cTotal*cTotal + sTotal*sTotal;
      }
      mmax(pkSumFlux, 1+2*nGrad, &s, &c, &ix, &iGrad);
/*      iGrad = nGrad - iGrad; ???? */
      iGrad = nGrad + iGrad; 

      cTotal = 0;
      sTotal = 0;
      for (ix=rmin; ix<rmax;ix++) {
         dPhase = iGrad * phasePerFrame * (ix-iRow);
         s      = sin(dPhase);
         c      = cos(dPhase);

         cTotal = cTotal + totalFlux[2*ix]*c - totalFlux[2*ix+1]*s;
         sTotal = sTotal + totalFlux[2*ix]*s + totalFlux[2*ix+1]*c;
      } 
      phase[iRow] = - atan2(sTotal, cTotal);
/*
 *                    try to unbias phase and then rotate it out
*/
      for (ix=0; ix<2*nx; ix++) tmpspec[ix] = dataArray[iRow]->data[0][ix];
      for (ix=0; ix<nx; ix++) {
         cTotalC = cTotal;
         sTotalC = sTotal;
         xmin = ix-NDEBIAS;  if (xmin < 0)  xmin =  0;
         xmax = ix+NDEBIAS; if (xmax >= nx) xmax = nx-1;
         c    = dataArray[iRow]->data[0][2*ix];
         s    = dataArray[iRow]->data[0][2*ix+1];
         for (idb=xmin;idb<=xmax;idb++) {
            cTotalC = cTotalC - tmpspec[2*idb];
            sTotalC = sTotalC - tmpspec[2*idb+1];
         }
         w = sqrt(cTotalC*cTotalC + sTotalC*sTotalC);
         dataArray[iRow]->data[0][2*ix]   = (c*cTotalC + s*sTotalC)/w;
         dataArray[iRow]->data[0][2*ix+1] = (s*cTotalC - c*sTotalC)/w;
      }
/*                    end of debias */
   }
   if (isMask) for (iRegion=0;iRegion<nRegion;iRegion++) free(maskImages[iRegion]);
   free(totalFlux);
} /* end of findRemovePhases */

double interpolateGroupDelay(double time, long nDelay, double *groupDelay, double *delayTime, long *lastFound) {

   long  iDelay;
   float x;

   iDelay = *lastFound;
   if (iDelay < 0) iDelay = 0;

   while (((iDelay < (nDelay-1)) && (delayTime[iDelay+1] < time)) || isnan(delayTime[iDelay+1]) || isnan(delayTime[iDelay])) ++iDelay;
   x = (time-delayTime[iDelay])/(delayTime[iDelay+1]-delayTime[iDelay]);
   *lastFound = iDelay;
   return (1-x)*groupDelay[iDelay] + x*groupDelay[iDelay+1];
} /*  end of interpolateGroupDelay */
int main (int argc, char** argv) {

   int  i, j, ierr;
   int  iy, y1, y2;
   long nRow, iRow, oRow, pRow ;
   int  nRegion, iRegion;
   long nDelay;
   long lastFound;

   int  ix;
   int  nx, ny;
   int  xMask, yMask;
   long npix, ii;
   float *frequency, *waveNo;
   double opd, iopd;


   char outFileName[130];
   char *bang="!";

   fitsfile *inFile;
   fitsfile *outFile ;
   fitsfile *phaseFile ;

   float *maskImages[12];
   int   isMask;

   int iKey, maxstep;

   imaging_fdata_table_def *idata;
   imaging_fdata_table_def *odata;
   imaging_detector_table_def *outDetectorDef;
   imaging_fdata_row *inData;
   imaging_fdata_row **outDataArray, *outData;

   delay_table_row *delayData;
   delay_table_def *dTable;
   double* groupDelay;
   double* delayTime;
  
   
   float *phase, c,s,w;

   float detdit;
   float sigma;
   int   doFrames;

   int doOpd, doPhase, doGroup;

   dispersion_table_def *disTable;
   dispersion_table_row *disRow;
   char* phaseFileName = NULL;
   char* maskFileName  = NULL;
   int   iarg    = 4;
   int   nGrad   = 0;
   outFile       = NULL;
   phaseFile     = NULL;
   ierr          = 0;
   sigma         = SMOOTHSIGMA;
/*                  for debugging allow me to turn things off */
   doPhase       = 1;
   doOpd         = 1;
   doGroup       = 1;
   doFrames      = 0;  /* smooth  specified in seconds by default */
   if (argc < 3) {
      Usage();
      return 1;
   }
   while (iarg < argc ) {
      if (!strcmp("-phase", argv[iarg]))  phaseFileName = argv[++iarg];
      if (!strcmp("-mask", argv[iarg]))   maskFileName  = argv[++iarg];
      if (!strcmp("-nophase",argv[iarg])) doPhase       = 0; /* hidden debugging option */
      if (!strcmp("-noopd",argv[iarg]))   doOpd         = 0; /* hidden debugging option */
      if (!strcmp("-nogroup",argv[iarg])) doGroup       = 0; /* hidden debugging option */
      if (!strcmp("-frames",argv[iarg]))  doFrames      = 1; /* smooth specified in frames*/
      if (!strcmp("-smooth", argv[iarg])) sscanf(argv[++iarg],"%f", &sigma);
      if (!strcmp("-ngrad", argv[iarg]))  sscanf(argv[++iarg],"%i", &nGrad);
      ++iarg;
   }
   printf("      ******   oirRotateGroupDelay   ******\n");
   printf("   input fringe file:             %s\n", argv[1]);
   printf("   output fringe file:            %s\n", argv[3]);
   printf("   group delay information from:  %s\n", argv[2]);
   if (phaseFileName == NULL) 
      printf("   Deriving phase information derived from the data.\n");
      else printf("   phase information from:        %s\n", phaseFileName);
   if (maskFileName == NULL) printf("   no mask used\n");
   printf("   excluded debias frames       = %i\n", 2*NDEBIAS+1);


/*
      open delay file and get delay data
*/
   fits_open_file (&inFile, argv[2], READONLY, &ierr);
   if (ierr != 0) {
      printf("error %i opening delay file %s\n", ierr, argv[2]);
      printf("%s \n",strerror(errno));
      return 1;
   }
   dTable = open_delay_table(inFile, &ierr);
   if (ierr != 0) {
      printf("error %i opening delay table \n", ierr);
      return 1;
   }
   delayData = make_delay_table_row(dTable);
   i= fits_get_num_rows(inFile, &nDelay, &ierr);
   groupDelay = (double *)malloc(nDelay*sizeof(double));;
   delayTime  = (double *)malloc(nDelay*sizeof(double));;
   for (iRow =1; iRow<=nDelay;iRow++) {
      i=read_delay_table(inFile, dTable, delayData, iRow, &ierr);
      if (ierr != 0) {
         printf("error %i reading delay table row %li\n", ierr, iRow);
         return 1;
      }
      groupDelay[iRow-1] = delayData->delay*clight; /*meters*/
      delayTime[iRow-1]  = delayData->time; /*meters*/
   }
/*
      close file and clean up structures
*/
   i = close_delay_table(inFile, dTable, &ierr);
   kill_delay_table_row(delayData);
/*
      open input file 
*/
   fits_open_file (&inFile, argv[1], READONLY, &ierr);
   if (ierr != 0) {
      printf("error %i opening data file %s\n", ierr, argv[1]);
      printf("%s \n",strerror(errno));
      return 1;
   }
/*
 *    get integration time
*/
   i = fits_read_key(inFile, TFLOAT, "HIERARCH ESO DET DIT", &detdit,0,&ierr);
   printf("   exposure time (DIT)          = %8.3f sec\n", detdit);
   if (doFrames) sigma = sigma * detdit;
   printf("   phase smoothing sigma        = %8.3f sec ==%8.1f frames\n", sigma, sigma/detdit);
   printf("   number of linear phase fits  = %i\n",2*nGrad+1);
   if (!doGroup) printf("   NOT applying Group Delay phase correction.\n");
   if (!doOpd)   printf("   NOT applying Instrumental Delay phase correction.\n");
   if (!doPhase) printf("   NOT applying water vapor phase correction.\n");

   if (detdit > 1.e-4) sigma = round(sigma/detdit);
/*
      create output data table 
*/
   strcpy(outFileName, bang);
   strcat(outFileName, argv[3]);

   fits_create_file (&outFile, outFileName, &ierr);
   if (ierr !=0) {
      fprintf (stderr, "error opening fits output %s %i\n",argv[3], ierr);
      printf("%s \n",strerror(errno));
      return 1;
      }

/*
      copy header from input->output
*/
   i = fits_copy_header(inFile, outFile, &ierr);

/*
      get detector information
*/
   outDetectorDef = read_imaging_detector_table(inFile, &ierr);
   if (ierr != 0) {
      printf("error %i reading detector table \n", ierr);
      return 1;
   }
   idata          = open_imaging_fdata_table(inFile, &ierr);
   if (ierr != 0) {
      printf("error %i reading data table \n", ierr);
      return 1;
   }
   iKey = table_find_key(idata->def, "MAXSTEP");
   if (iKey>0)  table_get_keyword (idata->def, iKey, (void*)&maxstep,
      &ierr);

/*     
   get row size, then double it (pseudo-complex)
*/
   nRegion = outDetectorDef->nregion;
   nx      = outDetectorDef->naxis[0][0];
   ny      = outDetectorDef->naxis[0][1];
   npix    = (long)nx * ny;
   for (iRegion=0;iRegion<nRegion;iRegion++)
      outDetectorDef->naxis[iRegion][0] = 2*nx; 

/*
      create output table definitions
*/

   odata = make_imaging_fdata_table_def (outFile, 
      outDetectorDef, maxstep, &ierr); 
   if (ierr !=0) {
      fprintf (stderr, "error creating data definition\n");
      return 1;
      }
   inData = make_imaging_fdata_row(idata);
   i = fits_get_num_rows (inFile,  &nRow, &ierr); 
   outDataArray = (imaging_fdata_row **) 
      malloc (nRow * sizeof(imaging_fdata_row *));
   for (i=0;i<nRow;i++) outDataArray[i] = make_imaging_fdata_row(odata);
/*       
      get frequency information from detector table
*/
   frequency = imaging_detector_wcs(outDetectorDef, 1);
   waveNo    = (float *)malloc(nx * sizeof(float));
   for (i=0; i<nx; i++) waveNo[i] = twopi * frequency[i]/clight;
   phase       = (float *)malloc(nRow * sizeof(float));
/*
 *        if phase specified externally, go get it
*/
   if (phaseFileName != NULL) {
      fits_open_file (&phaseFile, phaseFileName, READONLY, &ierr);
      if (ierr != 0) {
         printf("error %i opening phase file %s\n", ierr, phaseFileName);
         printf("%s \n",strerror(errno));
         return 1;
      }
      disTable = open_dispersion_table(phaseFile, &ierr);
      if (ierr != 0) {
         printf("error %i opening phase table in %s\n", ierr, phaseFileName);
         printf("%s \n",strerror(errno));
         return 1;
      }
      disRow = make_dispersion_table_row(disTable);
      for (iRow=1; iRow<=nRow; iRow++) {
         i = read_dispersion_table(phaseFile, disTable, disRow,
            iRow, &ierr);
         if (ierr != 0) {
            printf("error %i reading phase table %s at row %li of %li\n", ierr, phaseFileName, iRow,nRow);
            printf("%s \n",strerror(errno));
            return 1;
         }
         phase[iRow-1] = -disRow->dispersion/radeg; 
      }
      kill_dispersion_table_def(disTable);
      kill_dispersion_table_row(disRow);
      i = fits_close_file(phaseFile, &ierr);
   }
/*       
         loop over input data, put in cyclic smoothing buffer,
         multiplying input value * exp(i*k*(D-opd0))
*/
   lastFound = -1;  /* initialize interpolation pointer */
   for (iRow=1; iRow<=nRow; iRow++) {
/*
        read new row, multiply by cos/sin of phase and store in output structure
*/
      i = read_imaging_fdata_table(inFile, idata, inData, iRow, &ierr);
      if (ierr != 0) {
         printf("error %i reading data table row %li\n", ierr, iRow);
         return 1;
      }
      outData = outDataArray[iRow-1];

      outData->time         = inData->time;
      outData->opd[0]       = inData->opd[0];
      outData->opd[1]       = inData->opd[1];
      outData->local_opd[1] = inData->local_opd[1];
      outData->local_opd[0] = inData->local_opd[0];
      outData->exp          = inData->exp;
      outData->offset[0]    = inData->offset[0];
      outData->offset[1]    = inData->offset[1];
      outData->rotation     = inData->rotation;
      outData->frame        = inData->frame;
      outData->step         = inData->step;
      outData->reference    = inData->reference;
      outData->targtype[0]  = inData->targtype[0];  
/*
         calculate opd, remove group delay, rotate fringe
*/
      opd = 0.;

      if (doOpd) opd = inData->opd[1]+inData->local_opd[1] - 
         inData->opd[0] - inData->local_opd[0];
      if (doGroup) {
         if (!isnan(inData->time)) {
            iopd = interpolateGroupDelay(outData->time, nDelay,
               groupDelay, delayTime, &lastFound);
            opd  = opd - iopd;
         } 
      }
      for (j=0; j<nx; j++) {
         c = cos(opd * waveNo[j]);
         s = sin(opd * waveNo[j]);
         for (iRegion=0;iRegion<nRegion;iRegion++) {
            for (iy=0;iy<ny;iy++) {
               ii = j+iy*nx;
               outData->data[iRegion][2*ii]   = inData->data[iRegion][ii]* c;
               outData->data[iRegion][2*ii+1] = inData->data[iRegion][ii]* s;
            }
         }
      }
   } /* all input data has been read in and rotated */
   i = close_imaging_fdata_table(inFile, idata, &ierr);
   kill_imaging_fdata_row(inData);
/*
      if no external phase table was specified, figure them out yourself 
      otherwise use specified table
*/
   if (doPhase) {
      if (phaseFileName == NULL) findRemovePhases(nRegion, outDataArray, nRow, nx, ny, 
         phase, sigma, maskFileName, nGrad, &ierr);

      else for (iRow=1;iRow<=nRow;iRow++) {
         rotateDataPhase(nRegion, outDataArray[iRow-1]->data, 
         nx, ny, phase[iRow-1]);
      }
   } 
   for (iRow=1; iRow<=nRow;iRow++) {
      i = write_imaging_fdata_table(outFile, odata, outDataArray[iRow-1], 
         iRow, &ierr);
      if (ierr != 0) {
         printf("error %i writing data table row %li\n", ierr, iRow);
         return 1;
      }
      if (0 == iRow % 100) fprintf(stderr, "Frame %ld of %ld\r",iRow, nRow);
   }

/*
 *    write found phases into output data table 
 */

   if (phaseFileName == NULL) {
      disTable = make_dispersion_table_def(outFile, "1776-07-04",
         "Constant Phase", &ierr);
      disRow = make_dispersion_table_row(disTable);
      for (iRow=1; iRow<=nRow; iRow++) {
         disRow->time       = outDataArray[iRow-1]->time;
         disRow->telescope  = 1;
         disRow->dispersion = -phase[iRow-1]*radeg;
         write_dispersion_table(outFile, disTable, disRow, iRow, &ierr);
      }
      kill_dispersion_table_def(disTable);
      kill_dispersion_table_row(disRow);
   }
   for (i=0; i< nRow; i++) kill_imaging_fdata_row(outDataArray[i]);
   kill_imaging_fdata_table_def(odata);
   i = fits_close_file(outFile, &ierr);
   free(frequency);
   free(waveNo);
   free (groupDelay);
   free (delayTime);
   free(phase);
   free(outDataArray);
   printf("");
   return 0;
}
